/*
 * MCONS
 *
 * multiple client program for the OSU/CIS console server
 *
 * This program will spy on multiple server consoles, giving a summary
 * display of thier activity on one terminal.  It assumes an ANSI
 * compatible color display, and allows the user to choose the text
 * attribute, and background and foreground colors for each server to
 * be monitored.
 *
 * Written 10/12/94 by Lane Robert (lar@usl.edu)
 * Modified 12/01/95 to filter nuisance escape sequences (Pyramid)
 * Modified 01/03/95 to allow reconnection attempts when connection drops
 * Modified 08/02/95 to clean up error messages and improve error checking
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <termio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <pwd.h>
#include <unistd.h>

struct esc_filter_state {
     char esc_seen;
     char skipping_chars;
     int chars_to_skip;
};

struct server_info {
     char *server_name;
     char *conserv_hostname;
     short conserv_portnum;
     char *conserv_password;
     time_t retry_at;
     int console_fd;
     char *ansi_mode_str;
     char *linebuf;
     char *lineptr;
     struct esc_filter_state fs;
     struct server_info *next;
};
char *me;
int tty_cols;
int verbose = 0;

extern int errno;
extern char *sys_errlist[];
extern char *optarg;

struct server_info *parse_config_file(char *);
int connect_to_server(char *, char *, short, char *);
void usage(void);
void readline(int, char *, int);

main(int argc, char **argv)
{
     char *conf_name;
     struct winsize ttywin;
     struct timeval timeout;
     char ch;
     struct server_info *servlist, *curnode;
     fd_set rfds;
     int nfiles, len, nc, i;

     conf_name = "/etc/mconstab";
     me = argv[0];

     /* read and save the terminal width */
     if (ioctl(1, TIOCGWINSZ, &ttywin) < 0) {
	  fprintf(stderr,
		  "%s: Unable to read terminal dimensions.  Assuming 80x24.\n",
		  me);
	  tty_cols = 80;
     }
     else tty_cols = (int)ttywin.ws_col;

     /* parse the command line arguments */
     while ((ch = getopt(argc, argv, "vf:")) != -1) {
	  switch (ch) {
	  case 'v':
		verbose++;
		break;
	  case 'f':
	       conf_name = optarg;
	       break;
	  default:
	       usage();
	       exit(-1);
	  }
     }

     /* read the config file, and make connections to the server */
     if ((servlist = parse_config_file(conf_name)) == NULL)
	  exit(-1);

     putchar('\n');

     while (1) {
	  FD_ZERO(&rfds);
	  curnode = servlist;
	  while (curnode != NULL) {
	       FD_SET(curnode -> console_fd, &rfds);
	       curnode = curnode -> next;
	  }

          timeout.tv_sec = 60;  /* wakeup every 60 secs */
          timeout.tv_usec = 0;
	  nfiles = select(8 * sizeof(rfds), &rfds, NULL, NULL, &timeout);
	  if (nfiles < 0) {
               perror("Select failed...continuing anyhow.\n");
	       continue;
          }

	  curnode = servlist;
	  while (curnode != NULL) {
               if (curnode -> console_fd == -1) {
                    if(time(NULL) >= curnode -> retry_at) {
			 curnode -> console_fd = 
			   connect_to_server(curnode -> server_name,
					     curnode -> conserv_hostname,
					     curnode -> conserv_portnum,
					     curnode -> conserv_password);
                         curnode -> retry_at = time(NULL) + 120;  /* in 2 mins */
			 printf("%smcons: Reconnecting to %s\033[m\n", 
				curnode -> ansi_mode_str,
				curnode -> server_name);
                    }
               }
	       else if (FD_ISSET(curnode -> console_fd, &rfds)) {
		    if (read(curnode -> console_fd, &ch, 1) < 1) { /* EOF (!!) or err */
			 close(curnode -> console_fd);
			 curnode -> console_fd = -1;
                         curnode -> retry_at = time(NULL) + 120;  /* in 2 mins */
		    }
		    else if (iscntrl(ch)) {
			 switch (ch) {
			 case '\b':
			      if (curnode -> lineptr > curnode -> linebuf)
				   curnode -> lineptr--;
			      break;
			 case '\t':
			      if ((len = curnode -> lineptr -
				         curnode -> linebuf) < tty_cols)
				   *(curnode -> lineptr++) = ' ';
			      len++;
			      while (len < tty_cols && len++ % 8)
				   *(curnode -> lineptr++) = ' ';
			      break;
			 case '\n':
			      nc = esc_filter(curnode -> linebuf,
					      strlen(curnode -> linebuf),
					      &(curnode -> fs));
			      curnode -> linebuf[nc] = '\0';
			      /* make ctrl chars visible */
#if 0
			      for (i = 0; i < nc; i++)
				   curnode -> linebuf[i] ^= 0100;
#endif
			      printf("%s%s\033[0m\n", curnode -> ansi_mode_str,
				     curnode -> linebuf);
			      memset(curnode -> linebuf, 0, tty_cols + 1);
			      curnode -> lineptr = curnode -> linebuf;
			      break;
			 case '\r':
			      curnode -> lineptr = curnode -> linebuf;
			      break;
			 default:
			      if (curnode -> lineptr - curnode -> linebuf <
				  tty_cols)
				   *(curnode -> lineptr)++ = ch;
			      break;
			 }
		       }
		    else if (curnode -> lineptr - curnode -> linebuf < tty_cols)
			 *(curnode -> lineptr)++ = ch;
	       }
	       curnode = curnode -> next;
	  }
     }
}
				   

/*
 * parse_config_file
 *
 * Open a given file, parse lines of the form:
 *   server:console-host:console-port:password:ansi-attrib:ansi-bg:ansi-fg
 * and create a linked list of server_info structures.  Returns NULL and
 * sets errno on error.
 */
struct server_info *
parse_config_file(char *conf_name)
{
     FILE *conf_fp;
     struct server_info *list, *newnode;
     char buf[256], *server_name, *console_host, *password, *ptr;
     int lineno, ansi_attrib, ansi_bg, ansi_fg;
     short console_port;

     if (conf_name == NULL) {
	  errno = EINVAL;
	  return NULL;
     }
     if ((conf_fp = fopen(conf_name, "r")) == NULL)
	  return NULL;
     lineno = 0;
     list = NULL;
     while (fgets(buf, sizeof(buf), conf_fp) != NULL) {
	  lineno++;
	  if (buf[0] == '#')
	       continue;
	  server_name = strtok(buf, ":");
	  console_host = strtok(NULL, ":");
	  ptr = strtok(NULL, ":");
	  if (ptr != NULL && isdigit(ptr[0]))
	       console_port = (short)atoi(ptr);
	  else console_port = -1;
	  password = strtok(NULL, ":");
	  ptr = strtok(NULL, ":");
	  if (ptr != NULL && isdigit(ptr[0]))
	       ansi_attrib = atoi(ptr);
	  else ansi_attrib = -1;
	  ptr = strtok(NULL, ":");
	  if (ptr != NULL && isdigit(ptr[0]))
	       ansi_bg = atoi(ptr);
	  else ansi_bg = -1;
	  ptr = strtok(NULL, ":");
	  if (ptr != NULL && isdigit(ptr[0]))
	       ansi_fg = atoi(ptr);
	  else ansi_fg = -1;
	  if (ptr == NULL || console_port <= 0){
	       fprintf(stderr, "%s: %s: Line %d is malformed. Skipping\n",
		       me, conf_name, lineno);
	       continue;
	  }
	  if (ansi_attrib < 0 || ansi_attrib > 8) {
	       fprintf(stderr,
		       "%s: %s: Bad ANSI attribute at line %d.  Using 0\n",
		       me, conf_name, lineno);
	       ansi_attrib = 0;
	  }
	  if (ansi_bg < 0 || ansi_bg > 7) {
	       fprintf(stderr,
		     "%s: %s: Bad ANSI background color at line %d. Using 0.\n",
		       me, conf_name, lineno);
	       ansi_bg = 0;
	  }
	  if (ansi_fg < 0 || ansi_fg > 7) {
	       fprintf(stderr,
		    "%s: %s: Bad ANSI foreground color at line %d: Using 7.\n",
		       me, conf_name, lineno);
	       ansi_fg = 7;
	  }
	  if ((newnode = malloc(sizeof(struct server_info))) == NULL)
	       return NULL;
	  newnode -> fs.esc_seen = 0;
	  newnode -> fs.skipping_chars = 0;
	  if ((newnode -> server_name = malloc(strlen(server_name) + 1)) ==
	      NULL)
	       return NULL;
	  strcpy(newnode -> server_name, server_name);
	  if ((newnode -> conserv_hostname = malloc(strlen(console_host) + 1))
	      == NULL)
	       return NULL;
	  strcpy(newnode -> conserv_hostname, console_host);
	  newnode -> conserv_portnum = console_port;
	  if ((newnode -> conserv_password =
	       malloc(strlen(password) + 1)) == NULL)
	       return NULL;
	  strcpy(newnode -> conserv_password, password);
	  if ((newnode -> ansi_mode_str = malloc(11)) == NULL)
	       return NULL;
	  sprintf(newnode -> ansi_mode_str, "\033[%d;3%d;4%dm", ansi_attrib,
		  ansi_fg, ansi_bg);
	  if ((newnode -> linebuf = malloc(tty_cols + 1)) == NULL)
	       return NULL;
	  memset(newnode -> linebuf, 0, tty_cols + 1);
	  if ((newnode -> console_fd =
	       connect_to_server(server_name, console_host, console_port,
				 password)) < 0)
	       return NULL;
	  newnode -> lineptr = newnode -> linebuf;
	  newnode -> next = list;
	  list = newnode;
     }
     fclose(conf_fp);

     return(list);
}
	  

/*
 * connect_to_server
 *
 * Returns the file descriptor of a socket connected to the
 * console server of a given server, in spy mode.
 */

int
connect_to_server(char *server_name, char *conserv_hostname,
		  short conserv_portnum, char *password)
{
     struct sockaddr_in conserver;
     struct hostent *conserv_addr;
     struct passwd *mypwd;
     char buf[128], myhost[32];
     int sock;

     if (verbose) fprintf(stderr, "Connecting to %s (for %s)...\n", conserv_hostname, server_name);

     /* Verify arguments */
     if (server_name == NULL || conserv_hostname == NULL ||
	 password == NULL || conserv_portnum < 0) {
	  fprintf(stderr, "%s: connect_to_server(): invalid args\n", me);
	  return -1;
     }
     /* Find address of console server */
     if ((conserv_addr = gethostbyname(conserv_hostname)) == NULL) {
	  fprintf(stderr, "%s: name lookup of %s failed: %s\n", me,
		  conserv_hostname, sys_errlist[errno]);
	  return -1;
     }
     memset(&conserver, 0, sizeof(conserver));
     conserver.sin_family = AF_INET;
     memcpy(&conserver.sin_addr, conserv_addr -> h_addr,
	    conserv_addr -> h_length);
     conserver.sin_port = htons(conserv_portnum);
     /* Create a stream socket and connect it to console server */
     if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  fprintf(stderr, "%s: connect_to_server(): cannot create socket: %s\n",
		  me, sys_errlist[errno]);
	  return -1;
     }
     if (connect(sock, &conserver, sizeof(conserver)) != 0) {
	  close(sock);
	  fprintf(stderr, "%s: connection to host %s master port failed: %s\n",
		  me, conserv_hostname, sys_errlist[errno]);
	  return -1;
     }
     /* Send the name of the desired server */
     sprintf(buf, "%s\n", server_name);
     write(sock, buf, strlen(buf));
     /* Server should respond with a port number, or "Sorry" */
     readline(sock, buf, sizeof(buf));
     close(sock);
     if (!isdigit(buf[0])) {
	  /* Server not found, so bail out */
	  fprintf(stderr, "%s: server %s not found\n", me, server_name);
	  return -1;
     }
     /* Connect to the returned port on the console server
      * and send the command to start spy mode
      */
     conserver.sin_port = htons((short)atoi(buf));
     mypwd = getpwuid(getuid());
     gethostname(myhost, sizeof(myhost));
     sprintf(buf, "%s:spy:%s:%s:%s\n", server_name,
	     mypwd ? mypwd -> pw_name : "unknown", myhost, password);
     if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  fprintf(stderr, "%s: connect_to_server(): cannot create socket: %s\n",
		  me, sys_errlist[errno]);
	  return -1;
     }
     if (connect(sock, &conserver, sizeof(conserver)) != 0) {
	  close(sock);
	  fprintf(stderr, "%s: connection to host %s port %d failed: %s\n",
		  me, conserv_hostname, ntohs(conserver.sin_port),
		  sys_errlist[errno]);
	  return -1;
     }
     write(sock, buf, strlen(buf));
     /* After sending the startup command, the server
      * should respond with "ok", so read that here.
      */
     readline(sock, buf, sizeof(buf));

     if (strncmp(buf, "ok", 2) != 0) {
	  fprintf(stderr, "%s: connection for server %s failed: bad password\n",
		  me, server_name);
	  close(sock);
	  return -1;
     }

     return sock;
}

void
usage(void)
{
     fprintf(stderr, "%s: Usage: \"%s [-v] [-f <config_file>]\"\n", me, me);
}

void
readline(int fd, char *buf, int len)
{
     char *ptr, ch;

     ptr = buf;
     read(fd, &ch, 1);
     while (ch != '\n' && ptr - buf < len) {
	  if (ch != '\r')
	       *ptr++ = ch;
	  read(fd, &ch, 1);
     }
     *ptr = '\0';

     return;
}

/* Filter out nuisance escape sequences.
 * Return the new length of the buffer.
 * Stores state information in *state to handle escape
 * sequences that span buffers.
 * 11/94 lar@usl.edu
 */
int
esc_filter(buf, nch, state)
char *buf;
int nch;
struct esc_filter_state *state;
{
     int i, start;

     start = 0;
     for (i = 0; i < nch; i++) {
     redo:
	  if (state -> esc_seen) {
	       state -> esc_seen = 0;
	       if (buf[i] == 'z') {
		    state -> chars_to_skip = 9;
		    state -> skipping_chars = 1;
		    continue;
	       }
	  }
	  if (state -> skipping_chars) {
	       if (state -> chars_to_skip == 0) {
		    memcpy(&buf[start], &buf[i], nch - i);
		    state -> skipping_chars = 0;
		    nch -= i - start;
		    i = start;
		    goto redo;
	       }
	       else if (buf[i] == '\r')
		         state -> chars_to_skip = 0;
	            else state -> chars_to_skip--;
	       continue;
	  }
	  if (buf[i] == '\033') {
	       state -> esc_seen = 1;
	       start = i;
	  }
     }
     if (state -> skipping_chars)
	  nch -= nch - start;

     return nch;
}
